#include "gtypes.h"
#include "gfx.h"

/* ------------------------------------------------ */
/* Lock and unlock surface before puting pixel into */
/* ------------------------------------------------ */

void LockSurface(SDL_Surface *screen){
	if ( SDL_MUSTLOCK(screen) ) {
    	  if ( SDL_LockSurface(screen) < 0 ) {
            return;
        }
    }
}

void UnlockSurface(SDL_Surface *screen){
	if ( SDL_MUSTLOCK(screen) ) {
		   SDL_UnlockSurface(screen);
	}
}


/* -------------------------- */
/* Computing of Madelbrot Set */
/* -------------------------- */
void MandelbrotSet(SDL_Surface *surface, unsigned int width, unsigned int height, 
		RealPlane Plane, unsigned int MaxIter, 
		T_ColorMap ColorMap)
{
/* http://en.wikipedia.org/wiki/Mandelbrot_set */
	unsigned int row, col, color, i = 0;
	double c_imag, c_real, Z_imag, Z_real, Z2_imag, Z2_real;
	const double DeltaX = (Plane.XMax - Plane.XMin) / (double) width;
	const double DeltaY = (Plane.YMax - Plane.YMin) / (double) height;

	LockSurface(surface);
	/* Mandelbrot set computation */
	c_imag = Plane.YMin;	
	for(row = 0; row < height; row++){
		c_real = Plane.XMin;
		for(col = 0; col < width; col++){
			color = 0;
			Z_real = 0.0;
			Z_imag = 0.0;
			
			/* Iteration */
			do {
			
				/* Z_{n+1} = Z_n^2 + c */
				/* (A + Bi)^2 = A^2 + 2*A*Bi - B^2 */

				Z2_real = Z_real * Z_real; 
				Z2_imag = Z_imag * Z_imag;
				Z_imag = 2.0 * Z_imag * Z_real + c_imag;
				Z_real = Z2_real - Z2_imag + c_real;	
				color++;

			} while(color < MaxIter && (Z2_real + Z2_imag) < 4.0L);
			
			
			/* plot pixel */
			
			/* surface->pixel -- pointer to image memory */
			ColorMap(surface, i, color, MaxIter);
			/*
			if(color==MaxIter)
				*((Uint32 *) surface->pixels + i) = 0x000000FF;
			else
				*((Uint32 *) surface->pixels + i) = SDL_MapRGBA(surface->format, (Uint8) color*10, (Uint8) color*9, (Uint8) color*4, 0xFF );
				*/


			i++;
			c_real += DeltaX;  
		}
		c_imag += DeltaY;  
	}		
	UnlockSurface(surface);

}

/* ---------------------- */
/* Computing of Julia Set */
/* ---------------------- */
void JuliaSet(SDL_Surface *surface, unsigned int width, unsigned int height, 
		RealPlane Plane, cnum c, unsigned int MaxIter, 
		T_ColorMap  ColorMap)
{
/* http://en.wikipedia.org/wiki/Julia_Set */
/* c -- starting point in complex plane choosen from Mandelbrot set */
	unsigned int row, col, color, i = 0;
	double x, y;
	register double Z_imag;
	register double Z_real;
	register double Z2_imag;
	register double Z2_real;
	const double c_imag = c.imag;
	const double c_real = c.real;
	const double DeltaX = (Plane.XMax - Plane.XMin) / (double) width;
	const double DeltaY = (Plane.YMax - Plane.YMin) / (double) height;
	
	LockSurface(surface);
	/* Julia set computation */
	y = Plane.YMin;
	for(row = 0; row < height; row++){
		x = Plane.XMin;
		for(col = 0; col < width; col++){
			color = 0;
			Z_real = x; /* Z := x+yi */
			Z_imag = y;
			
			/* Iteration */
			do {
				/* Z_{n+1} = Z_n^2 + c */
				Z2_real = Z_real * Z_real; 
				Z2_imag = Z_imag * Z_imag;
				Z_imag = 2.0 * Z_imag * Z_real + c_imag;
				Z_real = Z2_real - Z2_imag + c_real;	
				color++;
			} while(color < MaxIter && (Z2_real + Z2_imag) < 4.0L);

			
			/* plot pixel */
			ColorMap(surface, i, color, MaxIter);
			/* surface->pixel -- pointer to image memory */
/*
			if(color==MaxIter)
				*((Uint32 *) surface->pixels + i) = black; 
			else
				*((Uint32 *) surface->pixels + i) = SDL_MapRGBA(surface->format, (Uint8) color*12, (Uint8) color*14, (Uint8) color*4, 0xFF );
			*/
			i++;
			x += DeltaX;  
		}
		y += DeltaY;  
	}		
	UnlockSurface(surface);
}

/* ----------------------------------------------------- */
/* Return pixel data in surface from actual x,y position */
/* ----------------------------------------------------- */
Uint32 GetPixel(SDL_Surface *surface, Uint16 X, Uint16 Y){
	Uint32 pixel;

	switch(surface->format->BytesPerPixel)
	{
		case 1: { /* 8-bit */
					pixel = *(Uint8 *)((Uint8 *) surface->pixels + Y * surface->pitch + X);
				}
			break;
		case 2: { /* 15-bit / 16-bit */
					pixel = *(Uint16 *)((Uint16 *) surface->pixels + Y * surface->pitch/2 + X);
				}
			break;
		case 3: { /* 24-bit */
					/* Not supported */
				}
			break;
		case 4: { /* 32-bit */
					pixel = *(Uint32 *)((Uint32 *) surface->pixels + Y * surface->pitch/4 + X);
				}
			break;
	}
/*
   {
	Uint8 r,g,b;
	SDL_GetRGB(pixel, surface->format, &r, &g, &b); 
	printf("\n(%d) 0x%08x ...  r = %u, g = %u, b = %u\n", surface->format->BytesPerPixel, pixel, r,g,b); 
	}
*/
	return pixel;
}

/* --------------------------- */
/* Draw singel pixel on screen */
/* --------------------------- */

/*
   Use this function after LockSurface() and do SDL_UpdateRect(...) after 
   finish of pixel works */

void DrawPixel(SDL_Surface *screen, int x, int y, Uint8 R, Uint8 G, Uint8 B, Uint8 A){
    Uint32 color = SDL_MapRGBA(screen->format, R, G, B, A);
   
	switch (screen->format->BytesPerPixel) {
        case 1: { /* Assuming 8-bpp */
            Uint8 *bufp;

            bufp = (Uint8 *)screen->pixels + y*screen->pitch + x;
            *bufp = color;
        }
        break;

        case 2: { /* Probably 15-bpp or 16-bpp */
            Uint16 *bufp;

            bufp = (Uint16 *)screen->pixels + y*screen->pitch/2 + x;
            *bufp = color;
        }
        break;

        case 3: { /* Slow 24-bpp mode, usually not used */
            Uint8 *bufp;

            bufp = (Uint8 *)screen->pixels + y*screen->pitch + x;
            *(bufp+screen->format->Rshift/8) = R;
            *(bufp+screen->format->Gshift/8) = G;
            *(bufp+screen->format->Bshift/8) = B;
        }
        break;

        case 4: { /* Probably 32-bpp */
            Uint32 *bufp;

            bufp = (Uint32 *)screen->pixels + y*screen->pitch/4 + x;
            *bufp = color;
        }
        break;
    }
/*    SDL_UpdateRect(screen, x, y, 1, 1);  */
}

void DrawZoomArea(SDL_Surface *surface, int XPress, int YPress, int XCurrent, int YCurrent){ 
	int x, y;

	/* Draw rectangle */

	if(XCurrent > XPress){
		for(x = XPress; x <= XCurrent; x++){	/* top line */
			DrawPixel(surface, x, YPress, ZOOM_RECT_COLOR);
			/* bottom line */
			DrawPixel(surface, x, YCurrent, ZOOM_RECT_COLOR);
		}
	}
	else {
		for(x = XPress; x >= XCurrent; x--){
			/* top line */
			DrawPixel(surface, x, YPress, ZOOM_RECT_COLOR);
			/* bottom line */
			DrawPixel(surface, x, YCurrent, ZOOM_RECT_COLOR);
		}
	}

	if(YCurrent > YPress){
		for(y = YPress; y <= YCurrent; y++ ){ 
			/* left line */
			DrawPixel(surface, XPress, y, ZOOM_RECT_COLOR);
			/* right line */
			DrawPixel(surface, XCurrent, y, ZOOM_RECT_COLOR);
		}
	}
	else{
		for(y = YPress; y >= YCurrent; y-- ){ /* left line */
			DrawPixel(surface, XPress, y, ZOOM_RECT_COLOR);
            /* right line */
			DrawPixel(surface, XCurrent, y, ZOOM_RECT_COLOR);
		}
	}


}

/* --------------------------------------- */
/* Functions for differnt palletes drawing */
/* --------------------------------------- */

/* Array with funtion T_ColorMap pointers */
T_ColorMap Array_ColorMap[NUMBER_COLOR_MAP] = { 
	PalGray, 
	PalBW, 
	PalColorGreen, 
	PalColorRed, 
	PalColorDarkBlue, 
	PalColorRainbow,
	PalColorLightBlue
};

void PalGray(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF ); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,
				(Uint8) iterations*10, (Uint8) iterations*10, (Uint8) iterations*10, 0xFF );
}

void PalBW(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF ); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,  0xFF,  0xFF,  0xFF,  0xFF);
}

void PalColorGreen(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,
				(Uint8) iterations*10, (Uint8) iterations*15, (Uint8) iterations*5, 0xFF );
}

void PalColorRed(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,
				(Uint8) iterations*17, (Uint8) iterations*3, (Uint8) iterations*3, 0xFF );
}

void PalColorDarkBlue(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,
				(Uint8) iterations*3, (Uint8) iterations*5, (Uint8) iterations*8, 0xFF );
}

void PalColorRainbow(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,
				(Uint8) iterations*30, (Uint8) iterations*50, (Uint8) iterations*80, 0xFF );
}

void PalColorLightBlue(SDL_Surface *surface, unsigned int index, unsigned int iterations, unsigned int MaxIter){
	if(iterations == MaxIter)
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format, 0, 0, 0, 0xFF); 
	else
		*((Uint32 *) surface->pixels + index) = SDL_MapRGBA(surface->format,
				(Uint8) iterations*5, (Uint8) iterations*5, (Uint8) iterations*20, 0xFF );
}/* -------------------------------------- */
